# -----------------------------------------
# Endroid - XMPP Bot
# Copyright 2012, Ensoft Ltd.
# Created by Jonathan Millican
# -----------------------------------------

import logging
from twisted.words.protocols.jabber.jid import JID


class Handler(object):
    __slots__ = ("name", "priority", "callback")
    def __init__(self, priority, callback):
        self.name = callback.__name__
        self.priority = priority
        self.callback = callback

    def __str__(self):
        return "{}: {}".format(self.priority, self.name)

class Priority(object):
    NORMAL = 0
    URGENT = -1
    BULK = 1

class MessageHandler(object):
    """An abstraction of XMPP's message protocols."""

    PRIORITY_NORMAL = Priority.NORMAL
    PRIORITY_URGENT = Priority.URGENT
    PRIORITY_BULK = Priority.BULK

    def __init__(self, wh, um):
        self.wh = wh
        self.um = um
        # wh translates messages and gives them to us, needs to know who we are
        self.wh.set_message_handler(self)
        self._handlers = {}

    def _register_callback(self, name, typ, cat, callback,
                           including_self=False, priority=Priority.NORMAL):
        """
        Register a function to be called on receipt of a message of type 
        'typ' (muc/chat), category 'cat' (recv, send, unhandled, *_self, *_filter)
        sent from user or room 'name'.

        """
        # self._handlers is a dictionary of form:
        # { type : { category : { room/groupname : [Handler objects]}}}
        typhndlrs = self._handlers.setdefault(typ, {})
        cathndlrs = typhndlrs.setdefault(cat, {})
        handlers = cathndlrs.setdefault(name, [])
        handlers.append(Handler(priority, callback))
        handlers.sort(key=lambda h: h.priority)

        # this callback be called when we get messages sent by ourself
        if including_self:
            self._register_callback(name, typ, cat + "_self", callback,
                                    priority=priority)

    def _get_handlers(self, typ, cat, name):
        dct = self._handlers.get(typ, {}).get(cat, {})
        if typ == 'chat':  # we need to lookup name's groups
            # we may have either a full jid or just a userhost,
            # groups are referenced by userhost
            name = self.um.get_userhost(name)
            handlers = []
            for name in self.um.get_groups(name):
                handlers.extend(dct.get(name, []))
            handlers.sort(key=lambda h: h.priority)
            return handlers
        else:  # we are in a room so only one set of handlers to read
            return dct.get(name, [])

    def _get_filters(self, typ, cat, name):
        return self._get_handlers(typ, cat + "_filter", name)

    def _do_callback(self, cat, msg, failback=lambda m: None):
        if msg.place == "muc":
            # get the handlers active in the room - note that these are already
            # sorted (sorting is done in the register_callback method)
            handlers = self._get_handlers(msg.place, cat, msg.recipient)
            filters = self._get_filters(msg.place, cat, msg.recipient)
        else: 
            # combine the handlers from each group the user is registered with
            # note that if the same plugin is registered for more than one of
            # the user's groups, the plugin's instance in each group will be
            # called
            handlers = self._get_handlers(msg.place, cat, msg.sender)
            filters = self._get_filters(msg.place, cat, msg.sender)
   
        log_list = []
        if handlers and all(f.callback(msg) for f in filters):
            msg.set_unhandled_cb(failback)

            for i in handlers:
                msg.inc_handlers()

            log_list.append("Did {} {} handlers (priority: cb):".format(len(handlers), cat))
            for handler in handlers:
                try:
                    handler.callback(msg)
                    log_list.append(str(handler))
                except Exception as e:
                    log_list.append("Exception in {}:\n{}".format(handler.name, e))
                    msg.dec_handlers()
                    raise
        else:
            failback(msg)
        if log_list:
            logging.info("Finished plugin callback: {}".format(
                         "\n\t".join(log_list)))
        else:
            logging.info("Finished plugin callback - no plugins called.")

    def _unhandled(self, msg):
        self._do_callback("unhandled", msg)

    def _unhandled_self(self, msg):
        self._do_callback("unhandled_self", msg)

    # Do normal (recv) callbacks on msg. If no callbacks handle the message
    # then call unhandled callbacks (msg's failback is set self._unhandled_...
    # by the last argument to _do_callback).
    def receive_muc(self, msg):
        self._do_callback("recv", msg, self._unhandled)

    def receive_self_muc(self, msg):
        self._do_callback("recv_self", msg, self._unhandled_self)

    def receive_chat(self, msg):
        self._do_callback("recv", msg, self._unhandled)

    def receive_self_chat(self, msg):
        self._do_callback("recv_self", msg, self._unhandled_self)

    def for_plugin(self, pluginmanager, plugin):
        return PluginMessageHandler(self, pluginmanager, plugin)

    # API for global plugins

    def register(self, name, callback, priority=Priority.NORMAL, muc_only=False,
                 chat_only=False, include_self=False, unhandled=False,
                 send_filter=False, recv_filter=False):
        if sum(1 for i in (unhandled, send_filter, recv_filter) if i) > 1:
            raise TypeError("Only one of unhandled, send_filter or recv_filter "
                            "may be specified")
        if chat_only and muc_only:
            raise TypeError("Only one of chat_only or muc_only may be "
                            "specified")

        if unhandled:
            cat = "unhandled"
        elif send_filter:
            cat = "send_filter"
        elif recv_filter:
            cat = "recv_filter"
        else:
            cat = "recv"

        if not muc_only:
            self._register_callback(name, "chat", cat, callback,
                                    include_self, priority)
        if not chat_only:
            self._register_callback(name, "muc", cat, callback,
                                    include_self, priority)

    def send_muc(self, room, body, source=None, priority=Priority.NORMAL):
        """
        Send muc message to room.

        The message will be run through any registered filters before it is
        sent.

        """
        # Verify this is a room EnDroid knows about
        msg = Message('muc', source, body, self, recipient=room)
        # when sending messages we check the filters registered with the
        # _recipient_. Cf. when we receive messages we check filters registered
        # with the _sender_.
        filters = self._get_filters('muc', 'send', msg.recipient)

        if all(f.callback(msg) for f in filters):
            logging.info("Sending message to {}".format(room))
            self.wh.groupChat(JID(room), body)
        else:
            logging.info("Filtered out message to {}".format(room))

    def send_chat(self, user, body, source=None, priority=Priority.NORMAL):
        """
        Send chat message to person with address user.

        The message will be run through any registered filters before it is
        sent.

        """
        # Verify user is known to EnDroid
        msg = Message('chat', source, body, self, recipient=user)
        filters = self._get_filters('chat', 'send', msg.recipient)

        if all(f.callback(msg) for f in filters):
            logging.info("Sending message to {}".format(user))
            self.wh.chat(JID(user), body)
        else:
            logging.info("Filtered out message to {0}".format(user))


class PluginMessageHandler(object):
    """
    One of these exists per plugin, providing the API to handle messsages.
    """
    def __init__(self, messagehandler, pluginmanager, plugin):
        self._messagehandler = messagehandler
        self._pluginmanager = pluginmanager
        self._plugin = plugin

    def send_muc(self, body, source=None, priority=Priority.NORMAL):
        if self._pluginmanager.place != "room":
            raise ValueError("Not in a room")
        self._messagehandler.send_muc(self._pluginmanager.name, body,
                                      source=source, priority=priority)

    def send_chat(self, user, body, source=None, priority=Priority.NORMAL):
        if self._pluginmanager.place != "group":
            raise ValueError("Not in a group")
        if user not in self._pluginmanager.usermanagement.users(
                                                    self._pluginmanager.name):
            raise ValueError("Target user is not in this group")
        # Verify user is in the group we are in
        self._messagehandler.send_chat(user, body,
                                       source=source, priority=priority)

    def register(self, callback, priority=Priority.NORMAL, muc_only=False,
                 chat_only=False, include_self=False, unhandled=False,
                 send_filter=False, recv_filter=False):
        if self._pluginmanager.place == "room" and not chat_only:
            muc_only = True
        if self._pluginmanager.place == "group" and not muc_only:
            chat_only = True
        self._messagehandler.register(self._pluginmanager.name, callback,
                                      priority=priority, muc_only=muc_only,
                                      chat_only=chat_only,
                                      include_self=include_self,
                                      unhandled=unhandled,
                                      send_filter=send_filter,
                                      recv_filter=recv_filter)

class Message(object):
    def __init__(self, place, sender, body, messagehandler, recipient, handlers=0,
                 priority=Priority.NORMAL):
        self.place = place

        # sender_full is a string representing the full jid (including resource)
        # of the message's sender. Used in reply methods so that if a user is
        # logged in on several resources, the reply will be sent to the right
        # one
        self.sender_full = sender
        # a string representing the userhost of the message's sender. Used to
        # lookup resource-independant user properties eg their registered rooms
        self.sender = messagehandler.um.get_userhost(sender)
        self.body = body
        self.recipient = recipient

        # a count of plugins which will try to process this message
        self.__handlers = handlers
        self._messagehandler = messagehandler
        self.priority = priority

    def send(self):
        if self.place == "chat":
            self._messagehandler.send_chat(self.recipient, self.body, self.sender)
        elif self.place == "muc":
            self._messagehandler.send_muc(self.recipient, self.body, self.sender)

    def reply(self, body):
        if self.place == "chat":
            self._messagehandler.send_chat(self.sender_full, body, self.recipient)
        elif self.place == "muc":
            # we send to the room (the recipient), not the message's sender
            self._messagehandler.send_muc(self.recipient, body, self.recipient)

    def reply_to_sender(self, body):
        self._messagehandler.send_chat(self.sender_full, body, self.recipient)

    def inc_handlers(self):
        self.__handlers += 1

    def dec_handlers(self):
        self.__handlers -= 1
        self.do_unhandled()

    def unhandled(self, *args):
        """
        Notify the message that the caller hasn't handled it. This should only
        be called by plugins that have registered as a handler (and thus have
        incremented the handler count for this message).

        This method takes arbitrary arguments so it can be used as deferred
        callback or errback.

        """
        self.dec_handlers()

    def do_unhandled(self):
        if self.__handlers == 0 and hasattr(self, '_unhandled_cb'):
            self._unhandled_cb(self)

    def set_unhandled_cb(self, cb):
        self._unhandled_cb = cb
