# -----------------------------------------------------------------------------
# EnDroid - HTTP Interface
# Copyright 2012, Ensoft Ltd
# -----------------------------------------------------------------------------

import collections
import re
import logging

from endroid.pluginmanager import Plugin
from twisted.internet import reactor
from twisted.web.resource import Resource
from twisted.web.server import Site
from twisted.web.static import File

DEFAULT_HTTP_PORT = 8880
DEFAULT_HTTP_INTERFACE = '127.0.0.1'
DEFAULT_MEDIA_DIR = "/usr/share/endroid/media/"

NOT_FOUND_TEMPLATE = """
<html>
    <head><title>Handler not found</title></head>
    <body bgcolor="white">
        <h1>Handler not found</h1>
        Handler not found: %s
    </body>
</html>
"""

INDEX_TEMPLATE = """
<html>
    <head><title>EnDroid</title></head>
    <body bgcolor="white">
        <h1>EnDroid HTTP interface</h1>
        <img style='float: right;' src='_media/EnDroid.png'>
        <ul>
{}
        </ul>
    </body>
</html>"""

class HandlerNotFoundError(Exception):
    pass

class IndexPage(Resource):
    """
    Only registered resource. Route requests to callbacks.
    """
    
    def __init__(self, interface):
        Resource.__init__(self)
        self.interface = interface

    def getChild(self, name, request):
        if name == "_media":
            return self.interface._media
        else:
            return self
    
    def _render(self, request):
        try:
            callback = self.interface.lookup_handler(request.path)
        except HandlerNotFoundError as e:
            page = NOT_FOUND_TEMPLATE % e
        else:
            page = callback(request)
            
        return page

    render_GET = _render
    render_POST = _render


class HTTPInterfaceSingleton(object):
    """
    Start a webserver, allow callback registrations on URL paths, and route
    requests to callbacks.
    """

    def __init__(self):
        self.registrations = collections.defaultdict(list)

    def _index(self, request):
        return INDEX_TEMPLATE.format("\n".join("<li><a href='{0}/'>{0}</a></li>"
                                               .format(name)
                                               for name in self.registrations))

    def lookup_handler(self, path):
        """"
        Search for a callback function registered to handle the given path.

        Raises HandlerNotFoundError if no callback has been registered on the
        given path.
        """

        assert path[0] == '/'
        path = path[1:]

        spl = path.split('/', 1)
        if len(spl) != 2 and not spl[0]:
            return self._index
        if len(spl) != 2:
            raise HandlerNotFoundError("Path only has one component")
        plugin_name, sub_path = spl

        if plugin_name not in self.registrations:
            raise HandlerNotFoundError("Plugin %s has no registered paths" %
                                        plugin_name)

        for regex, cb in self.registrations[plugin_name]:
            if regex.match(sub_path):
                callback = cb
                break
        else:
            raise HandlerNotFoundError("Plugin %s is not registered on %s" %
                                            (plugin_name, sub_path))

        return callback
            
    def register_regex_path(self, plugin, callback, path_regex):
        """
        See HTTPInterface.register_regex_path()
        """
        if isinstance(path_regex, str):
            path_regex = re.compile(path_regex)

        self.registrations[plugin.name].append((path_regex, callback))

    def register_path(self, plugin, callback, path_prefix=''):
        """
        See HTTPInterface.register_path()
        """
        if not path_prefix:
            re_src = r".*"
        else:
            if path_prefix[0] == '/':
                raise ValueError("Prefix %s begins with a slash" % path_prefix)
            if path_prefix[-1] == '/':
                raise ValueError("Prefix %s ends with a slash" % path_prefix)
            re_src = re.escape(path_prefix) + r"(/.*)?"
        self.register_regex_path(plugin, callback, re_src)

    def register_resource(self, plugin, resource):
        self.root.putChild(plugin.name, resource)
        
    def endroid_init(self, port, interface, media_dir):
        self.root = IndexPage(self)
        logging.info("Publishing static files from {}".format(media_dir))
        self._media = File(media_dir)
        self.root.putChild("_media", self._media)
        factory = Site(self.root)
        reactor.listenTCP(port, factory, interface=interface)

class HTTPInterface(Plugin):
    """
    The actual plugin class. This may be instantiated multiple times, but is
    just a wrapper around a HTTPInterfaceSingleton object.
    """

    object = None
    enInited = False
    name = "httpinterface"
    hidden = True

    def __init__(self):
        if HTTPInterface.object == None:
            HTTPInterface.object = HTTPInterfaceSingleton()
    
    def register_regex_path(self, plugin, callback, path_regex):
        """
        Register a callback to be called for requests whose URI matches:
           http://<server>/<plugin name>/<regex>[?<args>]

        Callback arguments:
            request: A twisted.web.http.Request object.
        """

        HTTPInterface.object.register_regex_path(plugin, callback, path_regex)

    def register_path(self, plugin, callback, path_prefix):
        """
        Register a callback to be called for requests whose URI matches:
            http://<server>/<plugin name>/<prefix>[?<args>]

        Or:
            http://<server>/<plugin name>/<prefix>/<more path>[?<args>]

        Or if the prefix is the empty string:
            http://<server>/<plugin name>/<more path>[?<args>]
        """

        HTTPInterface.object.register_path(plugin, callback, path_prefix)

    def register_resource(self, plugin, resource):
        HTTPInterface.object.register_resource(self, plugin, resource)
        
    def endroid_init(self):
        if not HTTPInterface.enInited:
            port = self.vars.get("port", DEFAULT_HTTP_PORT)
            interface = self.vars.get("interface", DEFAULT_HTTP_INTERFACE)
            media_dir = self.vars.get("media_dir", DEFAULT_MEDIA_DIR)
            HTTPInterface.object.endroid_init(port, interface, media_dir)
            HTTPInterface.enInited = True
