# -----------------------------------------------------------------------------
# EnDroid - Memo Plugin
# Copyright 2012, Ben Eills
# -----------------------------------------------------------------------------

"""
Plugin to send email-like messages between users.
See MESSAGES['usage'] for details.
"""

from endroid.plugins.command import CommandPlugin, command
from endroid.database import Database
from endroid.database import EndroidUniqueID

# Constants for the DB values we use
DB_NAME = "Memo"
DB_TABLE = "Memos"
DB_COLUMNS = ("sender", "recipient", "text")
SUMMARY_WIDTH = 60
MESSAGES = {
    'help' : """
Memo is a system for sending short text messages.
Try 'memo usage' to see usage.""",

    'usage' : """
memo usage - Output this message
memo [list] - List messages in inbox
memo send <recipient> <message> - Send a message
memo {view|delete} <message-id> - Handle a specific message""",

    'usage-send' : "memo send <recipient> <message> \t Send a message",
    'usage-view' : "memo view <message-id> \t View a specific message",
    'usage-delete' : "memo delete <message-id> \t Delete a specific message",

    'duplicate-id-error' : "Error: There are messages sharing the same PK id",
    'id-not-found-error' : "There is no message with id {0}.  Try 'memo list'.",
    'deletion-sucess' : "Sucessfully deleted one message",
    'inbox-header' : "Inbox:",
    'inbox-footer' : "Use 'memo view' to view a message.",
    'inbox-empty' : "Memo inbox empty.",
    'view-message' : "Message from {0}:\n{1}",
    'sent-message' : "Sucessfully sent message",
    'invalid-action' : "Invalid action : {0}",
    'bad-recipient': "Target recipient ({0}) is not known",
    }


class Memo(CommandPlugin):
    name = "memo"
    help_topics = {
        '': lambda _: MESSAGES['help'],
        'usage': lambda _: MESSAGES['usage'],
        'send': lambda _: MESSAGES['usage-send'],
        'view': lambda _: MESSAGES['usage-view'],
        'delete': lambda _: MESSAGES['usage-delete'],
    }
    
    def endroid_init(self):
        self.db = Database(DB_NAME)
        if not self.db.table_exists(DB_TABLE):
            self.db.create_table(DB_TABLE, DB_COLUMNS)

    def _message_text_summary(self, text):
        # First assume text is too long, try to shorten.
        summary = ""
        words = [w.strip() for w in text.split()]  # Get rid of newlines, etc.
        for word in words:
            if len(summary) + len(word) > SUMMARY_WIDTH - 3:
                return summary + '...'
            summary += word + ' '
        # Looks like memo is already short enough.  Return it.
        return summary[:-1]

    @command
    def memo_usage(self, msg, args):
        msg.reply(MESSAGES['usage'])

    @command(helphint="<message-id>")
    def memo_delete(self, msg, args):
        args = args.split()
        if len(args) != 1 or not args[0].isdigit():
            msg.reply(MESSAGES['usage-delete'])
            return
        cond = {EndroidUniqueID : int(args[0])}
        rows = self.db.fetch(DB_TABLE, DB_COLUMNS, cond)
        if len(rows) > 1:
            # SQLITE should guarantee this can never be reached, but
            #   let's be pedantic.
            msg.reply(MESSAGES['duplicate-id-error'].format(int(args[0])))
        elif len(rows) < 1:
            msg.reply(MESSAGES['id-not-found-error'])
        else:
            self.db.delete(DB_TABLE, {EndroidUniqueID : int(args[0])})
            msg.reply(MESSAGES['deletion-sucess'])

    @command(helphint="[list]")
    def memo(self, msg, args):
        args = args.split()
        if len(args) > 0 and args[0] in {"send", "view", "delete", "usage"}:
            msg.unhandled()
            return
        if len(args) > 1 or (len(args) == 1 and args[0] != "list"):
            msg.reply(MESSAGES['usage'])
            return
        memos = []
        cond = {"recipient" : msg.sender}
        rows = self.db.fetch(DB_TABLE, DB_COLUMNS, cond)
        for row in rows:
            data = (row.id, row["sender"],
                    self._message_text_summary(row["text"]))
            memos.append("[{0:5d}] [{1}]\t{2}".format(*data))
        if memos:
            msg.reply("{0}\n{1}\n{2}".format(MESSAGES['inbox-header'],
                                             '\n'.join(memos),
                                             MESSAGES['inbox-footer']))
        else:
            msg.reply(MESSAGES['inbox-empty'])
    
    @command(helphint="<message-id>")
    def memo_view(self, msg, args):
        args = args.split()
        if len(args) != 1 or not args[0].isdigit():
            msg.reply(MESSAGES['usage-view'])
            return
        cond = {EndroidUniqueID : int(args[0])}
        rows = self.db.fetch(DB_TABLE, DB_COLUMNS, cond)
        if len(rows) > 1:
            # SQLITE should guarantee this can never be reached, but
            #   let's be pedantic.
            msg.reply(MESSAGES['duplicate-id-error'].format(int(args[0])))
        elif len(rows) < 1:
            msg.reply(MESSAGES['id-not-found-error'])
        else:
            row = rows[0]
            data = (row["sender"], row["text"])
            msg.reply(MESSAGES['view-message'].format(*data))

    @command(helphint="<recipient> <message>")
    def memo_send(self, msg, args):
        args = args.split()
        if len(args) < 2:
            msg.reply(MESSAGES['usage-send'])
            return

        recipient, memo = args[0], ' '.join(args[1:])
        if recipient not in self.usermanagement.get_available_users():
            msg.reply(MESSAGES['bad-recipient'].format(recipient))
            return

        self.db.insert(DB_TABLE, {'sender':    msg.sender,
                                  'recipient': recipient,
                                  'text':      memo})
        msg.reply(MESSAGES['sent-message'])
