# -----------------------------------------
# Endroid - XMPP Bot
# Copyright 2012, Ensoft Ltd.
# Created by Jonathan Millican
# -----------------------------------------

import logging
from twisted.words.protocols.jabber.jid import JID, parse, InvalidFormat
from twisted.words.protocols.jabber.error import StanzaError
from endroid.pluginmanager import PluginManager
from random import choice
from collections import namedtuple

# we use ADJECTIVES to generate a random new nick if endroid's is taken
ADJECTIVES = [
    "affronted", "annoyed", "antagonized", "bitter", "chafed", "convulsed",
    "cross", "displeased", "enraged", "exasperated", "ferocious", "fierce",
    "fiery", "fuming", "furious", "galled", "hateful", "heated", "impassioned",
    "incensed", "indignant", "inflamed", "infuriated", "irascible", "irate",
    "ireful", "irritated", "maddened", "offended", "outraged", "provoked", 
    "raging", "resentful", "splenetic", "storming", "vexed", "wrathful"]

MUC = "muc#roomconfig_"

Place = namedtuple("Place", ("type", "name"))

class Roster(object):
    """
    Provides functions for maintaining sets of users registered with and
    available in a contact list, user group or room.

    """
    def __init__(self, name=None, registration_cb=None, deregistration_cb=None):
        self.name = name or "contacts"

        self._members = {}  # dict of userhosts to full jid strings
        self.registration_cb = registration_cb or (lambda a, b: None)
        self.deregistration_cb = deregistration_cb or (lambda a, b: None)

    @property
    def available(self):
        # jids is a set of the resources on which a user is avaialbe
        # if jids is empty then the user is unavailable 
        return set(u for (u, avs) in self._members.items() if avs)

    @property
    def registered(self):
        return set(self._members.keys())

    def get_resources(self, name):
        return self._members.get(name, {})

    def set_registration_list(self, names):
        # if we have a list callback then don't do sub-callbacks
        for name in self.registered:
            if not name in names:
                self.deregister_user(name)
        for name in names:
            self.register_user(name)

    def register_user(self, name):
        if not name in self._members:
            self._members[name] = {}
            self.registration_cb(name, self.name)

    def deregister_user(self, name):
        self._members.pop(name, None)
        self.deregistration_cb(name, self.name)

    def set_available(self, full_jid, show=None, priority=0):
        userhost = UserManagement.get_userhost(full_jid)
        logging.debug("{} came online as {}".format(userhost, full_jid))
        if userhost in self._members:
            self._members[userhost][full_jid] = Resource(show, priority)

    def set_unavailable(self, full_jid):
        userhost = UserManagement.get_userhost(full_jid)
        logging.debug("{} went offline as {}".format(userhost, full_jid))
        if userhost in self._members:
            self._members[userhost].pop(full_jid, None)

    def __repr__(self):
        name = self.name or "contacts"
        users = [u + ("(available)" if jids else "") 
                    for (u, jids) in self._members.items()]
        return "{}({}: {})".format(type(self).__name__, name, ', '.join(users))


class Room(Roster):
    config_default = {MUC + k : v for (k, v) in [
            ("allowinvites", False),
            ("changesubject",True),
            ("membersonly", True),
            ("moderatedroom", True),
            ("passwordprotectedroom", False),
            ("persistentroom", False),
            ("publicroom", False),
            ("roomdesc", "An Endroid chatroom"),
            ("roomname", "Ensoft"),
        ]
    }
    # these are the user configurable options
    config_options = {
        'password' : "roomsecret",
        'persistent' : "persistentroom",
        'name' : "roomname",
        'description' : "roomdesc",
    }

    @staticmethod
    def translate_options(options):
        def translate(key):
            return MUC + Room.config_options[key]

        do_password = options.get('password', False)
        options = {translate(k): v for (k, v) in options.items() 
                        if k in Room.config_options}
        if do_password:
            options[MUC + "passwordprotectedroom"] = True
        return options

    def __init__(self, name, nick, password, *args, **kwargs):
        self.nick = nick
        self.password = password
        # our add/remove user deferreds can only be run after we join the room,
        # so keep this list to store those generated before we do so
        self.deferreds = []
        super(Room, self).__init__(name, *args, **kwargs)

class Resource(object):
    __slots__ = ("show", "priority")

    def __init__(self, show=None, priority=0):
        self.show = show
        self.priority = priority

    def __repr__(self):
        name = type(self).__name__
        return "{0}(show='{1.show}', priority='{1.priority}'".format(name, self)

class UserManagement(object):
    """An abstraction of XMPP's presence protocols."""

    JOIN_ATTEMPTS_MAX = 5

    def __init__(self, wh, rh, config):
        self.wh = wh
        self.rh = rh
        # as yet unused - is analagous to the _handlers attribute of
        # messagehandler but will respond to presence notifications
        self._handlers = {}

        self._pms = {}  # a dict of {room/group names : pluginmanager objects}

        self.jid = None
        # our contact list and room list
        self._users = Roster(None, self._cb_add_contact, self._cb_rem_contact)
        self._rooms = Roster()
        # dictionaries of Roster objects for our groups and rooms
        self.group_rosters = {}
        self.room_rosters = {}

        self.conf = config
        self._read_config(config)

    def _read_config(self, config):
        def allowed_users(r_g, name):
            # get may return an empty string (if the line is 'users='), so use
            # or to make sure we have a set
            users = set(config.get(r_g, name, "users", default=self.users()))
            return users & self.users()

        # set our contact list and room list
        self.jid = config.get("setup", "jid")
        self._users.set_registration_list(config.get("setup", "users", default=[]))
        self._rooms.set_registration_list(config.get("setup", "rooms", default=[]))

        self.JOIN_ATTEMPTS_MAX = config.get("setup", "join_attempts", default=5)

        # set contact lists for our rooms
        for room in config.get("setup", "rooms", default=[]):
            # what we need to join the room
            nick = config.get("setup", "nick", default=parse(self.jid)[0])
            password = config.get("room", room, "plugin", 
                                  "endroid.plugins.roomowner", "password",
                                  default="")
            # user data
            self.room_rosters[room] = Room(room, nick, password, 
                                           self._cb_add_room_contact,
                                           self._cb_rem_room_contact)
            users = allowed_users("room", room)
            self.set_registration_list(users, room)

        for group in config.get("setup", "groups", default=['all']):
            self.group_rosters[group] = Roster(group)
            users = allowed_users("group", group)
            self.set_registration_list(users, group)

        # rosterhandler receives presence notifications and gives them to us
        self.rh.set_presence_handler(self)
        self.wh.set_presence_handler(self)

    # @@@ The below two function are unused - but show the structure that should
    # be used by register and call presence callbacks.
    def register_callback(self):
        # register functions to be called when we receive an 'online' or
        # 'offline' notifications
        pass

    def do_callback(self):
        pass

    ### Function about users and groups ###

    # given a group or room or None (our contact list), return list of users 
    # registered/available there
    def users(self, name=None):
        """
        Return an iterable of users registered with 'name'.

        If name is None, look in contact list.

        """
        if name is None:
            return self._users.registered
        elif name in self.group_rosters:
            return self.group_rosters[name].registered
        elif name in self.room_rosters:
            return self.room_rosters[name].registered
    get_users = users

    def resources(self, name=None):
        """
        Return an iterable of full jids at which 'name' is available.

        """
        if name is None:
            return self.jid
        else:
            return self._users.get_resources(name)

    def available_users(self, name=None):
        """
        Return an iterable of users present in 'name'.

        If name is None, look in contact list.

        """
        if name is None:
            return self._users.available
        elif name in self.group_rosters:
            return self.group_rosters[name].available
        elif name in self.room_rosters:
            return self.room_rosters[name].available
    get_available_users = available_users

    # given a user or None (us), return list of groups/rooms the user is 
    # registered/available in
    def groups(self, user=None):
        """
        Return an iterable of groups 'user' is registered with.

        If user is None, return all registered groups.

        """
        return self._get_user_place(user, self.group_rosters, get_available=False)
    get_groups = groups

    def available_groups(self, user=None):
        """
        Return an iterable of groups 'user' is present in.

        If user is None, return all groups EnDroid is available in.

        """
        return self._get_user_place(user, self.group_rosters, get_available=True)
    get_available_groups = available_groups

    def rooms(self, user=None):
        """
        Return an iterable of rooms 'user' is registered with.

        If user is None, return all registered rooms.

        """
        return self._get_user_place(user, self.room_rosters, get_available=False)
    get_rooms = rooms

    def available_rooms(self, user=None):
        """
        Return an iterable of rooms 'user' is present in.

        If user is None, return all rooms EnDroid is available in.

        """
        return self._get_user_place(user, self.room_rosters, get_available=True)
    get_available_rooms = available_rooms

    def _get_user_place(self, user, dct, get_available):
        """
        Return the list of places in dct 'user' is available in
        (if get_available) or registered with.

        """
        user = self.get_userhost(user)
        if user is None:
            return dct.keys()
        elif user in self.users():
            attr = "available" if get_available else "registered"
            return [p for p in dct.keys() if user in getattr(dct[p], attr)]
        else:
            return []

    def nickname(self, user, place=None):
        """
        Given a user jid (user@ho.s.t) return the user's nickname in place,
        or if place is None (default), the user part of the jid.

        """
        user = self.get_userhost(user)
        if place is None or place in self.group_rosters:
            return parse(user)[0]
        elif place in self.room_rosters:
            for (nick, rosteritem) in self.wh._rooms[JID(place)].roster.items():
                if user == rosteritem.entity.userhost():
                    return nick
        return "unknown"
    get_nickname = nickname

    ### Functions for managing contact lists ###

    def set_registration_list(self, names, place=None):
        """
        Set the 'member' list for our contacts (place = None) or in a group or
        room. This specifies who is allowed in said place.

        """
        if place is None:
            self._users.set_registration_list(names)
        elif place in self.group_rosters:
            self.group_rosters[place].set_registration_list(names)
        elif place in self.room_rosters:
            self.room_rosters[place].set_registration_list(names)

    def register_user(self, name, place=None):
        """
        Add a user to the 'member' list for our contacts (place=None) or in a
        group or room.

        """
        if place is None:
            self._users.register_user(name)
        elif place in self.group_rosters:
            self.group_rosters[place].register_user(name)
        elif place in self.room_rosters:
            self.room_rosters[place].register_user(name)

    def deregister_user(self, name, place=None):
        """
        Remove a user from the 'member' list for our contacts (place=None) or
        in a group or room.

        """
        if place is None:
            self._users.deregister_user(name)
        elif place in self.group_rosters:
            self.group_rosters[place].deregister_user(name)
        elif place in self.room_rosters:
            self.room_rosters[place].deregister_user(name)

    def set_available(self, name, place=None, show=None, priority=0):
        """
        Add a user to the 'available' list for our contacts (place=None) or in a
        group or room.

        """
        if place is None:
            self._users.set_available(name, show, priority)
        elif place in self.group_rosters:
            self.group_rosters[place].set_available(name, show, priority)
        elif place in self.room_rosters:
            self.room_rosters[place].set_available(name, show, priority)

    def set_unavailable(self, name, place=None):
        """
        Add a user to the 'available' list for our contacts (place=None) or in a
        group or room.

        """
        if place is None:
            self._users.set_unavailable(name)
        elif place in self.group_rosters:
            self.group_rosters[place].set_unavailable(name)
        elif place in self.room_rosters:
            self.room_rosters[place].set_unavailable(name)

    # callbacks for adding/removing from our main contact list
    def _cb_add_contact(self, name, _):
        logging.info("Adding contact: {}".format(name))
        self.rh.setItem(name)
        # send a subscribe request - confirmation in rosterhandler.subscribeReceived
        self.rh.subscribe(JID(name))
        self.rh.available(JID(name))

    def _cb_rem_contact(self, name, _):
        logging.debug("Removing contact: {}".format(name))
        self.rh.removeItem(name)  # sets the subscription to None both ways
        logging.info("Removed {} from {}".format(name, "contacts"))

    # callbacks for adding/removing from rooms - can only be called after we are
    # in the room ie if room is in self._wh.rooms. otherwise append the callback
    # to a list to be called when we join (from userJoinedRoom in wokkelhandler)
    def _cb_add_room_contact(self, name, room):
        fmt_add = "Added {} to {}"
        fmt_fail = "Failed to add {} to {}"

        if room in self.wh._rooms:
            d = self.wh.modifyAffiliationList(JID(room), [JID(name)], "member")
            d.addCallback(lambda _: logging.info(fmt_add.format(name, room)))
            d.addErrback(lambda _: logging.info(fmt_fail.format(name, room)))
            logging.debug("Adding {} to {}".format(name, room))
            return d
        else:
            self.room_rosters[room].deferreds.append((name, 1))

    def _cb_rem_room_contact(self, name, room):
        fmt_rem = "Removed {} from {}"
        fmt_fail = "Failed to remove {} from {}"

        if room in self.wh._rooms:
            d = self.wh.modifyAffiliationList(JID(room), [JID(name)], "none")
            d.addCallback(lambda _: logging.info(fmt_rem.format(name, room)))
            d.addErrback(lambda _: logging.info(fmt_fail.format(name, room)))
            logging.debug("Removing {} from {}".format(name, room))
            return d
        else:
            self.room_rosters[room].deferreds.append((name, 0))

    def _do_room_contacts(self, name):
        members, nones = [], []
        for (user, add) in self.room_rosters[name].deferreds:
            if add:
                members.append(user)
            else:
                nones.append(user)

        fmt_add = "Added {} to {}"
        fmt_rem = "Removed {} from {}"

        if members:
            string = fmt_add.format(', '.join(members), name)
            d1 = self.wh.modifyAffiliationList(JID(name), map(JID, members), "member")
            d1.addCallback(lambda _: logging.info(string))
        if nones:
            string = fmt_rem.format(', '.join(nones), name)
            d2 = self.wh.modifyAffiliationList(JID(name), map(JID, nones), "none")
            d2.addCallback(lambda _: logging.info(string))


    ### Room functions

    # @@@ at the moment this assumes that if we failed to join a room it was because
    # our nickname was taken - no provision for wrong passwords (should never be an
    # issue) or other failures
    def join_room(self, name, attempts=0):
        """
        Attempt to join the room 'name'. On failure will try again with
        different nicknames until attempts == JOIN_ATTEMPTS_MAX succes.

        """
        if not name in self.room_rosters:
            return "Failed to join room {}: not allowed".format(name)

        def retry(failure, room, nick, attempts):
            if attempts < self.JOIN_ATTEMPTS_MAX:
                return self.join_room(room, attempts+1)
            else:
                return "Failed to join room {} as {}".format(room, nick)
        def joined(_, room, nick, attempts):
            if attempts:
                self.wh.kick(JID(room), nick, "Begone thou who taketh my name!")
                self.wh.nick(JID(room), nick)
            return "Joined room {} as {}".format(room, nick)

        room = self.room_rosters[name]
        nick = room.nick
        if attempts:  # someone has stolen our nickname - generate a new one
            nick += " ({} and {})".format(choice(ADJECTIVES), choice(ADJECTIVES))

        d = self.wh.join(JID(name), nick, password=room.password)
        d.addErrback(retry, name, room.nick, attempts)
        d.addCallback(joined, name, room.nick, attempts)
        
        return d

    def joined_room(self, name):
        logging.info("Joined room {}".format(name))
        self._do_room_contacts(name)
        self._rooms.set_available(name)
        if not name in self._pms:
            self.start_pm(None, "room", name)

    def joined_group(self, name):
        logging.info("Initialised group {}".format(name))
        if not name in self._pms:
            self.start_pm(None, "group", name)

    # register a new pluginmanager (which will initialise plugins) in room or
    # group 'name'
    def start_pm(self, _, place, name):
        self._pms[name] = PluginManager(self.wh.messagehandler, self, place, 
                                        name, self.conf)

    def connected(self):
        self._pms[None] = PluginManager(self.wh.messagehandler, self, "global",
                                        None, self.conf)
        # Should join all rooms and groups here
        # Currently called from elsewhere

    def join_all_rooms(self):
        for room in self.get_rooms():
            d = self.join_room(room)
            d.addCallback(lambda _, r: self.joined_room(r), room)

    def join_all_groups(self):
        for group in self.get_groups():
            self.joined_group(group)


    def configure_room(self, name, options):
        """
        Set/modify the configuration of room 'name'. Options is a dictionary
        of option names to values. Currently configurable options 
        (Room.config_options) are: password, persistent, name (roomname) and 
        description.

        """
        def cb(value, name):
            try:
                failure = value.trap(StanzaError)
                if failure == StanzaError:
                    return "Configured {}: not all options set".format(name)
            except:
                return "Configured {}".format(name)

        mucoptions = Room.config_default.copy()
        # options is a human-readable dict, we intersect with Room.config_options
        # and translate into muc#roomconfig_xxx terms in translate_options
        mucoptions.update(Room.translate_options(options))
        d = self.wh.configure(JID(name), mucoptions)
        d.addBoth(cb, name)
        return d.addCallback(logging.info)

    def get_configuration(self, name):
        """
        Return the configuration of a room. This will return a list of strings
        representing option : value pairs in the configuration.

        """
        def form_to_string(form):
            items = ((k.replace(MUC, ""), v) for (k, v) in form.items())
            return '\n'.join("{} : {}".format(k, v) for (k, v) in items)

        if name in self.get_rooms():
            d = self.wh.getConfiguration(JID(name))
            return d.addCallback(form_to_string)

    def invite(self, user, room, reason=None):
        """
        Invite a user to a room.

        Will only send invitation if the user is in our contact list and online,
        and if the user is registered in the room but not currently in it.

        Returns a tuple (success, message)

        """
        user = self.get_userhost(user)
        if not user in self.users():
            return (False, "User not registered")
        elif not user in self.users():
            return (False, "User not available")
        elif not room in self.get_rooms(user):
            if not room in self.get_rooms():
                reason = "Room not registered"
            else:
                reason = "User not registered in room"
            return (False, reason)
        elif room in self.available_rooms(user):
            return (False, "User already in room")
        else:
            for full_jid in self.resources(user):
                self.wh.invite(JID(room), JID(full_jid), reason)
            return (True, "Invitation sent")

    @staticmethod
    def get_userhost(item):
        """
        This is usually applied to a Message's sender attribute which we expect
        to be a string or None. It will attempt to return a userhost string but
        if it fails will just return the input.

        """
        try:
            return JID(item).userhost()
            # will raise a RuntimeError if item is None
            # or an InvalidFormat if item is a string but not formatted properly
            # If the item is not even a string, will raise AttributeError
        except (RuntimeError, InvalidFormat, AttributeError):
            return item

    def for_plugin(self, pluginmanager, plugin):
        return PluginUserManagement(self, pluginmanager, plugin)

class PluginUserManagement(object):
    """
    One of these exists per plugin, provides the API to handle rosters.
    """
    def __init__(self, usermanagement, pluginmanager, plugin):
        self._usermanagement = usermanagement
        self._pluginmanager = pluginmanager
        self._plugin = plugin

    @property
    def users(self):
        return self._usermanagement.users(self._pluginmanager.name)

    @property
    def available_users(self):
        return self._usermanagement.available_users(self._pluginmanager.name)

    def nickname(self, user):
        return self._usermanagement.nickname(user, self._pluginmanager.name)

    def invite(self, user, reason=None):
        if self._pluginmanager.place != "room":
            raise ValueError("Must be in a room to invite users")
        return self._usermanagement.invite(user, self._pluginmanager.name,
                                           reason=reason)
