# -----------------------------------------
# Endroid - XMPP Bot
# Copyright 2012, Ensoft Ltd.
# Created by Jonathan Millican
# -----------------------------------------

from endroid.messagehandler import Message
from endroid.cron import Cron
import logging

from wokkel.muc import MUCClient
from wokkel.xmppim import MessageProtocol
from twisted.words.protocols.jabber.jid import JID
from twisted.internet.task import LoopingCall


# MUCClient provides muc messaging + room handling
# MessageProtocol provides suc messaging
class WokkelHandler(MUCClient, MessageProtocol):
    def __init__(self):
        MUCClient.__init__(self)
        MessageProtocol.__init__(self)

        self.messagehandler = None
        self.usermanagement = None
        self.deferred = None
        self.keepalive = LoopingCall(self.send, " ")

        self.jid = None

    def set_connected_handler(self, d):
        self.deferred = d

    def set_message_handler(self, mh):
        self.messagehandler = mh

    def set_presence_handler(self, um):
        self.usermanagement = um
        self.jid = self.usermanagement.jid

    # called when the xml stream has been initialized
    # use to do further initialization work
    def connectionInitialized(self):
        # adds MUCClientProtocol._onGroupChat to the xmlstream observers
        # notified by GROUPCHAT = "/message[@type="groupchat"]"
        MUCClient.connectionInitialized(self)
        # adds MessageProtocol._onMessage to the xmlstream observers
        # notified by "/message" i.e. anything
        MessageProtocol.connectionInitialized(self)
        self.keepalive.start(300)
        logging.info("Connected to XMPP")

        self.usermanagement.join_all_rooms()
        self.usermanagement.join_all_groups()


        # we start the cron here, then cancel it in connectionLost
        Cron.get().do_crons()

        if self.deferred:
            d, self.deferred = self.deferred, None
            d.callback(self)

    def connectionLost(self, reason):
        self.keepalive.stop()
        logging.warning("Lost connection to XMPP")
        Cron.get().cancel()


    # called by _onMessage whenever a stanza of type "/message" is received
    # we use it to pass the message onto our messagehandler
    def onMessage(self, message):
        MessageProtocol.onMessage(self, message)
        if message.hasAttribute('type'):
            if message.attributes['type'] == u'chat':
                body = None
                # we get the full jid string here so that if a user is logged
                # on from multiple resources, we know from which one the message
                # originated
                sender = JID(message.attributes['from']).full()
                recipient = JID(message.attributes['to']).full()
                for i in message.children:
                    # Note: in theory multiple bodies are allowed; in practice,
                    # this isn't seen, so just get the first one.
                    if getattr(i, "name", "") == 'body':
                        body = i.children[0]
                        break

                if not (body is None or sender is None or recipient is None):
                    m = Message("chat", sender, body, 
                                self.messagehandler, recipient)
                    if sender == self.jid:
                        logging.info('SUC Message received from self')
                        self.messagehandler.receive_self_chat(m)
                    else:
                        logging.info('SUC Message received from {}'.format(sender))
                        self.messagehandler.receive_chat(m)
        else:
            if message.children[0].children[0].name == "invite":
                ujid = message.children[0].children[0].attributes['from']
                rjid = message.attributes['from']
                logging.info("Invitation received to " + rjid + " from " + ujid)
                logging.info(("Rejected. To enable uncomment receiving "
                              "line and enable InviteBot"))

    # MUCClient methods to be overridden

    def userJoinedRoom(self, room, user):
        room_name = room.roomJID.userhost()
        user_name = user.entity.full()
        if user_name == self.jid:
            # This is EnDroid joining a room nothing to do
            pass
        else:
            logging.info(user_name + " joined " + room_name)
            self.usermanagement.set_available(user, room)

    def userLeftRoom(self, room, user):
        self.usermanagement.set_unavailable(user, room)

    def userUpdatedStatus(self, room, user, show, status):
        pass

    # new room subject received
    def receivedSubject(self, room, user, subject):
        pass

    def receivedGroupChat(self, room, user, message):
        MUCClient.receivedGroupChat(self, room, user, message)
        if user is None:
            logging.info('Received group chat with none user - '
                         'perhaps from a room? : {}'.format(message.body))
            return

        sender_jid = user.entity.full()
        room_userhost = room.roomJID.userhost()

        m = Message("muc", sender_jid, message.body, 
                    self.messagehandler, room_userhost)

        # If the jid specified in the config file has a resource,
        # then we need to reduce it to just the userhost string
        our_jid = self.usermanagement.get_userhost(self.jid)

        sender_nick = self.usermanagement.get_nickname(sender_jid, room_userhost)
        our_nick = self.usermanagement.get_nickname(our_jid, room_userhost)

        if sender_nick == our_nick:
            logging.info('MUC Message received from self')
            self.messagehandler.receive_self_muc(m)
        else:
            logging.info('MUC Message received from: ' + message.sender.full())
            self.messagehandler.receive_muc(m)

    def receivedHistory(self, room, user, message):
        pass
